/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 */

#include <qapplication.h>
#include <qmainwindow.h>
#include <qtoolbar.h>
#include <qvbox.h>
#include <qsplitter.h>
#include <qlistview.h>
#include <qtextview.h>
#include <qlineedit.h>
#include <qmenubar.h>
#include <qmessagebox.h>
#include <qaction.h>
#include <qheader.h>
#include <qfiledialog.h>
#include <stdlib.h>
#include "zconf.h"
#include "zconf.tab.h"
#include "qconf.h"
#include "version.h"

#include "qconf.moc"
#include "images.c"

/*
 * update all the children of a menu entry
 *   removes/adds the entries from the parent widget as necessary
 *
 * parent: either the menu list widget or a menu entry widget
 * menu: entry to be updated
 */
template <class P>
static void updateMenuList(P* parent, struct menu* menu, bool showAll, bool showFull)
{
	struct menu* child;
	ConfigItem* last;
	bool visible;

	last = 0;
	for (child = menu->list; child; child = child->next) {
		visible = menu_is_visible(child);

		if ((showAll || visible) && showFull) {
			if (!child->data)
				child->data = new ConfigItem(parent, last, child);
			last = (ConfigItem*)child->data;
			last->setEnabled(visible);
			last->updateMenu();
		} else {
			if (child->data) {
				struct menu* m;

				delete (ConfigItem*)child->data;
				// find the previous list entry
				last = 0;
				for (m = menu->list; m != child; m = m->next)
					if (m->data)
						last = (ConfigItem*)m->data;
			}
		}
	}
}

#if QT_VERSION >= 300
/*
 * set the new data
 * TODO check the value
 */
void ConfigItem::okRename(int col)
{
	Parent::okRename(col);
	sym_set_string_value(menu->sym, text(dataColIdx).latin1());
}
#endif

/*
 * update the displayed of a menu entry
 */
void ConfigItem::updateMenu(void)
{
	ConfigList* list = listView();
	struct symbol* sym;
	int type;
	tristate expr;
	bool showFull = true;

	if (menu->prompt && menu->prompt->token == T_MENU)
		showFull = list->showFull();

	sym = menu->sym;
	if (!sym)
		goto done;

	if (!(sym->flags & SYMBOL_CHANGED) && !list->updateAll)
		goto done;

	sym->flags &= ~SYMBOL_CHANGED;

	setText(nameColIdx, menu->sym->name);

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		char ch;

		setText(promptColIdx, QString().sprintf("%s%s", menu_get_prompt(menu),
				sym_has_value(sym) ? "" : " (NEW)"));
		if (!sym_is_changable(sym) && !list->showAll()) {
			setText(noColIdx, 0);
			setText(modColIdx, 0);
			setText(yesColIdx, 0);
			break;
		}
		expr = sym_get_tristate_value(sym);
		switch (expr) {
		case yes:
			if (sym_is_choice_value(sym) && type == SYMBOL_BOOLEAN)
				setPixmap(promptColIdx, list->choiceYesPix);
			else
				setPixmap(promptColIdx, list->symbolYesPix);
			setText(yesColIdx, "Y");
			ch = 'Y';
			break;
		case mod:
			setPixmap(promptColIdx, list->symbolModPix);
			setText(modColIdx, "M");
			ch = 'M';
			break;
		default:
			if (sym_is_choice_value(sym) && type == SYMBOL_BOOLEAN)
				setPixmap(promptColIdx, list->choiceNoPix);
			else
				setPixmap(promptColIdx, list->symbolNoPix);
			setText(noColIdx, "N");
			ch = 'N';
			break;
		}
		if (expr != no)
			setText(noColIdx, sym_tristate_within_range(sym, no) ? "_" : 0);
		if (expr != mod)
			setText(modColIdx, sym_tristate_within_range(sym, mod) ? "_" : 0);
		if (expr != yes)
			setText(yesColIdx, sym_tristate_within_range(sym, yes) ? "_" : 0);

#if 0
		if (type == SYMBOL_BOOLEAN)
			setText(promptColIdx, QString().sprintf("[%c] %s", ch, menu_get_prompt(menu)));
		else
			setText(promptColIdx, QString().sprintf("<%c> %s", ch, menu_get_prompt(menu)));
		setText(promptColIdx, menu_get_prompt(menu));
#endif
		setText(dataColIdx, QChar(ch));
		break;
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		const char* data;

#if 0
		if (S_VAL(sym->def))
			data = (char*)S_VAL(sym->def);
		else if (menu->prompt && menu->prompt->def) {
			sym_calc_value(menu->prompt->def);
			data = sym_get_string_value(menu->prompt->def);
		} else
#endif
			data = sym_get_string_value(sym);
#if QT_VERSION >= 300
		setRenameEnabled(list->mapIdx(dataColIdx), TRUE);
#endif
		setText(dataColIdx, data);
		if (type == SYMBOL_STRING)
			setText(promptColIdx, QString().sprintf("%s: %s", menu_get_prompt(menu), data));
		else
			setText(promptColIdx, QString().sprintf("(%s) %s", data, menu_get_prompt(menu)));
		break;
	}
done:
	updateMenuList(this, menu, list->showAll(), showFull);
}

/*
 * construct a menu entry
 */
void ConfigItem::init(void)
{
	ConfigList* list = listView();
#if QT_VERSION < 300
	setEnabled(TRUE);
#endif
	menu->data = this;
	if (menu->sym) {
		menu->sym->flags |= SYMBOL_CHANGED;
	} else {
		setText(promptColIdx, menu_get_prompt(menu));
		if (menu->prompt && menu->prompt->token == T_MENU && !list->showFull())
			setPixmap(promptColIdx, list->menuPix);
	}

	if (!list->showFull())
		setOpen(TRUE);
}

/*
 * destruct a menu entry
 */
ConfigItem::~ConfigItem(void)
{
	menu->data = 0;
}

void ConfigLineEdit::show(ConfigItem* i)
{
	item = i;
	if (sym_get_string_value(item->menu->sym))
		setText(sym_get_string_value(item->menu->sym));
	else
		setText(0);
	Parent::show();
	setFocus();
}

void ConfigLineEdit::keyPressEvent(QKeyEvent* e)
{
	switch (e->key()) {
	case Key_Escape:
		break;
	case Key_Return:
	case Key_Enter:
		sym_set_string_value(item->menu->sym, text().latin1());
		emit lineChanged(item);
		break;
	default:
		Parent::keyPressEvent(e);
		return;
	}
	e->accept();
	hide();
}

ConfigList::ConfigList(QWidget* p)
	: Parent(p),
	  updateAll(false),
	  symbolYesPix(xpm_symbol_yes), symbolModPix(xpm_symbol_mod), symbolNoPix(xpm_symbol_no),
	  choiceYesPix(xpm_choice_yes), choiceNoPix(xpm_choice_no), menuPix(xpm_menu), menuInvPix(xpm_menu_inv),
	  _showAll(false), _showName(true), _showRange(true), _showData(true), _showFull(true),
	  rootEntry(&rootmenu)
{
	int i;

	setSorting(-1);
	setRootIsDecorated(TRUE);

	for (i = 0; i < colNr; i++)
		colMap[i] = colRevMap[i] = -1;
	addColumn(promptColIdx, "prompt");

	//connect(this, SIGNAL(clicked(QListViewItem*, const QPoint&, int)),
	//	SLOT(setValue(QListViewItem*, const QPoint&, int)));

	reinit();
}

void ConfigList::reinit(void)
{
	removeColumn(dataColIdx);
	removeColumn(yesColIdx);
	removeColumn(modColIdx);
	removeColumn(noColIdx);
	removeColumn(nameColIdx);

	if (showName())
		addColumn(nameColIdx, "name");
	if (showRange()) {
		addColumn(noColIdx, "n");
		addColumn(modColIdx, "m");
		addColumn(yesColIdx, "y");
	}
	if (showData())
		addColumn(dataColIdx, "data");

	updateListAll();
}

void ConfigList::updateList(ConfigItem* item)
{
	(void)item;	// unused so far
	updateMenuList(this, rootEntry, showAll(), true);
}

void ConfigList::setValue(ConfigItem* item, tristate val)
{
	struct symbol* sym;
	int type;
	tristate oldval;

#if QT_VERSION < 300
	if (!item->enabled)
		return;
#endif
	sym = item->menu->sym;
	if (!sym)
		return;

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		oldval = sym_get_tristate_value(sym);

		if (!sym_set_tristate_value(sym, val))
			return;
		if (oldval == no && item->menu->list)
			item->setOpen(TRUE);
		updateList(item);
		break;
#if 0
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
		if (col == dataColIdx) {
#if QT_VERSION >= 300
			item->startRename(dataColIdx);
#else
			lineEdit->show(item);
#endif
		}
#endif
		break;
	}
}

void ConfigList::changeValue(ConfigItem* item)
{
	struct symbol* sym;
	struct menu* menu;
	int type, oldexpr, newexpr;

#if QT_VERSION < 300
	if (!item->enabled)
		return;
#endif

	menu = item->menu;
	sym = menu->sym;
	if (!sym) {
		if (item->menu->list)
			item->setOpen(!item->isOpen());
		return;
	}

	type = sym_get_type(sym);
	switch (type) {
	case SYMBOL_BOOLEAN:
	case SYMBOL_TRISTATE:
		oldexpr = sym_get_tristate_value(sym);
		newexpr = sym_toggle_tristate_value(sym);
		if (item->menu->list) {
			if (oldexpr == newexpr)
				item->setOpen(!item->isOpen());
			else if (oldexpr == no)
				item->setOpen(TRUE);
		}
		break;
	case SYMBOL_INT:
	case SYMBOL_HEX:
	case SYMBOL_STRING:
#if QT_VERSION >= 300
		if (colMap[dataColIdx] >= 0)
			item->startRename(colMap[dataColIdx]);
		else
#endif
			lineEdit->show(item);
		break;
	}

	updateList(item);
}

bool ConfigList::setRootMenu(struct menu *menu)
{
	if (!menu->prompt || menu->prompt->token != T_MENU)
		return false;
	updateMenuList(this, rootEntry, false, false);
	rootEntry = menu;
	updateListAll();
	return true;
}

void ConfigList::keyPressEvent(QKeyEvent* ev)
{
	QListViewItem* i = currentItem();
	ConfigItem* item;

	if (ev->key() == Key_Escape && !showFull() && rootEntry != &rootmenu) {
		struct menu *oldroot, *newroot;
		oldroot = newroot = rootEntry;
		do {
			newroot = newroot->parent;
			if (newroot->prompt && newroot->prompt->token == T_MENU)
				break;
		} while (newroot != &rootmenu);
		setRootMenu(newroot);
		setCurrentItem((ConfigItem*)oldroot->data);
		ev->accept();
		return;
	}

	if (!i) {
		Parent::keyPressEvent(ev);
		return;
	}
	item = (ConfigItem*)i;

	switch (ev->key()) {
	case Key_Return:
	case Key_Enter:
		if (!showFull() && setRootMenu(item->menu))
			break;
	case Key_Space:
		changeValue(item);
		break;
	case Key_N:
		setValue(item, no);
		break;
	case Key_M:
		setValue(item, mod);
		break;
	case Key_Y:
		setValue(item, yes);
		break;
	default:
		Parent::keyPressEvent(ev);
		return;
	}
	ev->accept();
}

void ConfigList::contentsMousePressEvent(QMouseEvent* e)
{
	//QPoint p(contentsToViewport(e->pos()));
	//printf("contentsMousePressEvent: %d,%d\n", p.x(), p.y());
	QListView::contentsMousePressEvent(e);
}

void ConfigList::contentsMouseReleaseEvent(QMouseEvent* e)
{
	QPoint p(contentsToViewport(e->pos()));
	ConfigItem* item = (ConfigItem*)itemAt(p);
	const QPixmap* pm;
	int idx, x;

	if (!item)
		goto skip;

	x = header()->offset() + p.x();
	idx = colRevMap[header()->sectionAt(x)];
	switch (idx) {
	case promptColIdx:
		pm = item->pixmap(promptColIdx);
		if (pm) {
			int off = header()->sectionPos(0) + itemMargin() +
				treeStepSize() * (item->depth() + (rootIsDecorated() ? 1 : 0));
			if (x >= off && x < off + pm->width()) {
				if (item->menu->sym)
					changeValue(item);
				else
					setRootMenu(item->menu);
			}
		}
		break;
	case noColIdx:
		setValue(item, no);
		break;
	case modColIdx:
		setValue(item, mod);
		break;
	case yesColIdx:
		setValue(item, yes);
		break;
	case dataColIdx:
		changeValue(item);
		break;
	}

	//printf("contentsMouseReleaseEvent: %d,%d\n", p.x(), p.y());
skip:
	QListView::contentsMouseReleaseEvent(e);
}

void ConfigList::contentsMouseMoveEvent(QMouseEvent* e)
{
	//QPoint p(contentsToViewport(e->pos()));
	//printf("contentsMouseMoveEvent: %d,%d\n", p.x(), p.y());
	QListView::contentsMouseMoveEvent(e);
}

void ConfigList::contentsMouseDoubleClickEvent(QMouseEvent* e)
{
	//QPoint p(contentsToViewport(e->pos()));
	//printf("contentsMouseDoubleClickEvent: %d,%d\n", p.x(), p.y());
	QListView::contentsMouseDoubleClickEvent(e);
}

void ConfigList::setShowAll(bool showAll)
{
	if (showAll == _showAll)
		return;
	_showAll = showAll;
	updateListAll();
	emit showAllChanged(showAll);
}

void ConfigList::setShowName(bool showName)
{
	if (showName == _showName)
		return;
	_showName = showName;
	reinit();
	emit showNameChanged(showName);
}

void ConfigList::setShowRange(bool showRange)
{
	if (showRange == _showRange)
		return;
	_showRange = showRange;
	reinit();
	emit showRangeChanged(showRange);
}

void ConfigList::setShowData(bool showData)
{
	if (showData == _showData)
		return;
	_showData = showData;
	reinit();
	emit showDataChanged(showData);
}

void ConfigList::setShowFull(bool showFull)
{
	if (showFull == _showFull)
		return;
	_showFull = showFull;
	setRootMenu(&rootmenu);
	emit showFullChanged(showFull);
}

/*
 * Construct the complete config widget
 */
ConfigView::ConfigView(void)
{
	QVBox *vbox;
	QMenuBar* menu;
	QSplitter* splitter;

	vbox = new QVBox(this);
	setCentralWidget(vbox);
	splitter = new QSplitter(vbox);
	splitter->setOrientation(QSplitter::Vertical);

	// create config tree
	QVBox* box = new QVBox(splitter);
	configList = new ConfigList(box);
	configList->lineEdit = new ConfigLineEdit(box);
	configList->lineEdit->hide();
	configList->connect(configList->lineEdit, SIGNAL(lineChanged(ConfigItem*)),
			    SLOT(updateList(ConfigItem*)));

	helpText = new QTextView(splitter);

	setTabOrder(configList, helpText);
	configList->setFocus();

	menu = menuBar();
	toolBar = new QToolBar("Tools", this);

	QAction *backAction = new QAction("Back", QPixmap(xpm_back), "Back", 0, this);
	  backAction->setEnabled(FALSE);
	QAction *quitAction = new QAction("Quit", "&Quit", CTRL+Key_Q, this);
	  connect(quitAction, SIGNAL(activated()), SLOT(close()));
	QAction *loadAction = new QAction("Load", QPixmap(xpm_load), "&Load", CTRL+Key_L, this);
	  connect(loadAction, SIGNAL(activated()), SLOT(loadConfig()));
	QAction *saveAction = new QAction("Save", QPixmap(xpm_save), "&Save", CTRL+Key_S, this);
	  connect(saveAction, SIGNAL(activated()), SLOT(saveConfig()));
	QAction *saveAsAction = new QAction("Save As...", "Save &As...", 0, this);
	  connect(saveAsAction, SIGNAL(activated()), SLOT(saveConfigAs()));

	QAction *fullTreeAction = new QAction(NULL, "Full Tree", 0, this);
	  fullTreeAction->setToggleAction(TRUE);
	  fullTreeAction->setOn(configList->showFull());
	  connect(fullTreeAction, SIGNAL(toggled(bool)), configList, SLOT(setShowFull(bool)));
	  //connect(configList, SIGNAL(showFullChanged(bool)), fullTreeAction, SLOT(setOn(bool)));
	QAction *showAllAction = new QAction(NULL, "Show All", 0, this);
	  showAllAction->setToggleAction(TRUE);
	  showAllAction->setOn(configList->showAll());
	  connect(showAllAction, SIGNAL(toggled(bool)), configList, SLOT(setShowAll(bool)));
	  //connect(configList, SIGNAL(showAllChanged(bool)), showAllAction, SLOT(setOn(bool)));
	QAction *showNameAction = new QAction(NULL, "Show Name", 0, this);
	  showNameAction->setToggleAction(TRUE);
	  showNameAction->setOn(configList->showName());
	  connect(showNameAction, SIGNAL(toggled(bool)), configList, SLOT(setShowName(bool)));
	  //connect(configList, SIGNAL(showNameChanged(bool)), showNameAction, SLOT(setOn(bool)));
	QAction *showRangeAction = new QAction(NULL, "Show Range", 0, this);
	  showRangeAction->setToggleAction(TRUE);
	  showRangeAction->setOn(configList->showRange());
	  connect(showRangeAction, SIGNAL(toggled(bool)), configList, SLOT(setShowRange(bool)));
	  //connect(configList, SIGNAL(showRangeChanged(bool)), showRangeAction, SLOT(setOn(bool)));
	QAction *showDataAction = new QAction(NULL, "Show Data", 0, this);
	  showDataAction->setToggleAction(TRUE);
	  showDataAction->setOn(configList->showData());
	  connect(showDataAction, SIGNAL(toggled(bool)), configList, SLOT(setShowData(bool)));
	  //connect(configList, SIGNAL(showDataChanged(bool)), showDataAction, SLOT(setOn(bool)));

	// init tool bar
	backAction->addTo(toolBar);
	toolBar->addSeparator();
	loadAction->addTo(toolBar);
	saveAction->addTo(toolBar);

	// create config menu
	QPopupMenu* config = new QPopupMenu(this);
	menu->insertItem("&File", config);
	loadAction->addTo(config);
	saveAction->addTo(config);
	saveAsAction->addTo(config);
	config->insertSeparator();
	quitAction->addTo(config);

	// create options menu
	QPopupMenu* optionMenu = new QPopupMenu(this);
	menu->insertItem("&Option", optionMenu);
	fullTreeAction->addTo(optionMenu);
	showAllAction->addTo(optionMenu);
	showNameAction->addTo(optionMenu);
	showRangeAction->addTo(optionMenu);
	showDataAction->addTo(optionMenu);

	connect(configList, SIGNAL(selectionChanged(QListViewItem*)),
		SLOT(setHelp(QListViewItem*)));
}

/*
 * display a new help entry as soon as a new menu entry is selected
 */
void ConfigView::setHelp(QListViewItem* item)
{
	struct symbol* sym;
	struct menu* menu;

	configList->lineEdit->hide();
	if (item) {
		menu = ((ConfigItem*)item)->menu;

		sym = menu->sym;
		if (sym) {
			helpText->setText(sym->help);
			return;
		}
	}
	helpText->setText(NULL);
}

void ConfigView::loadConfig(void)
{
	QString s = QFileDialog::getOpenFileName(conf_filename, NULL, this);
	if (s.isNull())
		return;
	if (conf_read(s.latin1()))
		QMessageBox::information(this, "qconf", "Unable to load configuration!");
}

void ConfigView::saveConfig(void)
{
	if (conf_write(NULL))
		QMessageBox::information(this, "qconf", "Unable to save configuration!");
}

void ConfigView::saveConfigAs(void)
{
	QString s = QFileDialog::getSaveFileName(conf_filename, NULL, this);
	if (s.isNull())
		return;
	if (conf_write(s.latin1()))
		QMessageBox::information(this, "qconf", "Unable to save configuration!");
}

/*
 * ask for saving configuration before quitting
 * TODO ask only when something changed
 */
void ConfigView::closeEvent(QCloseEvent* e)
{
	QMessageBox mb("qconf", "Save configuration?", QMessageBox::Warning,
			QMessageBox::Yes | QMessageBox::Default, QMessageBox::No, QMessageBox::Cancel | QMessageBox::Escape);
	switch (mb.exec()) {
	case QMessageBox::Yes:
		conf_write(NULL);
	case QMessageBox::No:
		e->accept();
		break;
	case QMessageBox::Cancel:
		e->ignore();
		break;
	}
}

int main(int ac, char** av)
{
	QApplication a(ac, av);
	ConfigView* v;
	const char *name;

	if (ac > 1 && av[1][0] == '-') {
		switch (av[1][1]) {
		case 'a':
			//showAll = 1;
			break;
		case 'h':
		case '?':
			printf("%s -a\n", av[0]);
			exit(0);
		}
		name = av[2];
	} else
		name = av[1];
	conf_parse(name);
	conf_read(NULL);
	//zconfdump(stdout);
	v = new ConfigView();

	//zconfdump(stdout);
	v->show();
	a.connect(&a, SIGNAL(lastWindowClosed()), &a, SLOT(quit()));
	a.exec();
	return 0;
}
